// SPDX-License-Identifier: MIT OR Apache-2.0

//! Console support protocols.
//!
//! The console represents the various input and output methods
//! used by the user to interact with the early boot platform.

pub mod gop;
pub mod pointer;
pub mod serial;
pub mod text;
